#! /usr/bin/env python
# coding=utf-8
import uuid

from twisted.trial import unittest

from pypegasus.pgclient import *
from pypegasus.rrdb.ttypes import filter_type
from pypegasus.utils.tools import MultiGetOptions


class TestBasics(unittest.TestCase):
    TEST_HKEY = 'test_hkey_1'
    TEST_SKEY = 'test_skey_1_'
    TEST_VALUE = 'test_value_1_'

    @inlineCallbacks
    def setUp(self):
        self.c = Pegasus(['127.0.0.1:34601', '127.0.0.1:34602', '127.0.0.1:34603'], 'temp')
        ret = yield self.c.init()
        self.assertTrue(ret)

    def tearDown(self):
        self.c.close()

    @inlineCallbacks
    def test_set_ok(self):
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE)
        self.assertEqual(ret, error_types.ERR_OK.value)

    @inlineCallbacks
    def test_set_timeout(self):
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE*1000000, 0, 10)
        self.assertEqual(ret, error_types.ERR_TIMEOUT.value)

    @inlineCallbacks
    def test_remove_ok(self):
        (ret, ign) = yield self.c.remove(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, v) = yield self.c.exist(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OBJECT_NOT_FOUND.value)

    @inlineCallbacks
    def test_exist_ok(self):
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, v) = yield self.c.exist(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OK.value)

    @inlineCallbacks
    def test_exist_none(self):
        (ret, ign) = yield self.c.remove(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, v) = yield self.c.exist(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OBJECT_NOT_FOUND.value)

    @inlineCallbacks
    def test_get_ok(self):
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, v) = yield self.c.get(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(v, self.TEST_VALUE)

    @inlineCallbacks
    def test_get_none(self):
        (ret, ign) = yield self.c.remove(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, v) = yield self.c.get(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OBJECT_NOT_FOUND.value)

    @inlineCallbacks
    def test_ttl_forever(self):
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, v) = yield self.c.ttl(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(v, -1)

    @inlineCallbacks
    def test_ttl_N(self):
        ttl = 60
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE, ttl)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, v) = yield self.c.ttl(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(v, ttl)

    @inlineCallbacks
    def test_ttl_N_with_phase(self):
        ttl = 10
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE, ttl)
        self.assertEqual(ret, error_types.ERR_OK.value)

        period = 2
        d = defer.Deferred()
        reactor.callLater(period, d.callback, 'ok')
        yield d

        (rc, v) = yield self.c.ttl(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(v, ttl - period)

    @inlineCallbacks
    def test_ttl_expired(self):
        ttl = 1
        (ret, ign) = yield self.c.set(self.TEST_HKEY, self.TEST_SKEY, self.TEST_VALUE, ttl)
        self.assertEqual(ret, error_types.ERR_OK.value)

        period = 1.5
        d = defer.Deferred()
        reactor.callLater(period, d.callback, 'ok')
        yield d

        (rc, v) = yield self.c.ttl(self.TEST_HKEY, self.TEST_SKEY)
        self.assertEqual(rc, error_types.ERR_OBJECT_NOT_FOUND.value)

    @inlineCallbacks
    def test_multi_set_ok(self):
        count = 50
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count)}
        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

    @inlineCallbacks
    def test_multi_get_ok(self):
        count = 50
        ks = {self.TEST_SKEY + str(x) for x in range(count)}
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, get_kvs) = yield self.c.multi_get(self.TEST_HKEY, ks)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(kvs))
        self.assertEqual(get_kvs, kvs)

    @inlineCallbacks
    def test_multi_get_opt_ok(self):
        count = 50
        rand_key = uuid.uuid1().hex
        kvs = {self.TEST_SKEY + rand_key + "_" + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        opt = MultiGetOptions()
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   self.TEST_SKEY + rand_key + '_0',
                                                   self.TEST_SKEY + rand_key + '_' + '9'*len(str(count)),
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(kvs))
        self.assertEqual(get_kvs, kvs)

    @inlineCallbacks
    def test_multi_get_opt_prefix_ok(self):
        count = 50
        rand_key = uuid.uuid1().hex
        rand_key2 = uuid.uuid4().hex
        kvs = {self.TEST_SKEY + rand_key + "_" + str(x): self.TEST_VALUE + str(x) for x in range(count)}
        kvs2 = {self.TEST_SKEY + rand_key2 + "_" + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)
        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs2)
        self.assertEqual(ret, error_types.ERR_OK.value)

        opt = MultiGetOptions()
        opt.sortkey_filter_type = filter_type.FT_MATCH_PREFIX
        opt.sortkey_filter_pattern = self.TEST_SKEY + rand_key
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   '',
                                                   '',
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(kvs))
        self.assertEqual(get_kvs, kvs)

    @inlineCallbacks
    def test_multi_get_opt_postfix_ok(self):
        count = 50
        rand_key = uuid.uuid1().hex
        rand_key2 = uuid.uuid4().hex
        kvs = {self.TEST_SKEY + str(x) + "_" + rand_key: self.TEST_VALUE + str(x) for x in range(count)}
        kvs2 = {self.TEST_SKEY + str(x) + "_" + rand_key2: self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)
        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs2)
        self.assertEqual(ret, error_types.ERR_OK.value)

        opt = MultiGetOptions()
        opt.sortkey_filter_type = filter_type.FT_MATCH_POSTFIX
        opt.sortkey_filter_pattern = rand_key
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   '',
                                                   '',
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(kvs))
        self.assertEqual(get_kvs, kvs)

    @inlineCallbacks
    def test_multi_get_opt_anywhere_ok(self):
        count = 50
        rand_key = uuid.uuid1().hex
        rand_key2 = uuid.uuid4().hex
        kvs = {self.TEST_SKEY + str(x) + "_" + rand_key + "_" + str(x): self.TEST_VALUE + str(x) for x in range(count)}
        kvs2 = {self.TEST_SKEY + str(x) + "_" + rand_key2 + "_" + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)
        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs2)
        self.assertEqual(ret, error_types.ERR_OK.value)

        opt = MultiGetOptions()
        opt.sortkey_filter_type = filter_type.FT_MATCH_ANYWHERE
        opt.sortkey_filter_pattern = rand_key
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   '',
                                                   '',
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(kvs))
        self.assertEqual(get_kvs, kvs)

    @inlineCallbacks
    def test_multi_get_opt_sortkey_inclusive_ok(self):
        rand_key = uuid.uuid1().hex
        start_key = self.TEST_SKEY + rand_key + '_start'
        stop_key = self.TEST_SKEY + rand_key + '_stop'
        start_value = self.TEST_VALUE + 'start'
        stop_value = self.TEST_VALUE + 'stop'
        kvs = {start_key: start_value,
               stop_key: stop_value
              }

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        opt = MultiGetOptions()

        # False False
        opt.start_inclusive = False
        opt.stop_inclusive = False
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   start_key,
                                                   stop_key,
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), 0)

        # False True
        opt.start_inclusive = False
        opt.stop_inclusive = True
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   start_key,
                                                   stop_key,
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), 1)
        self.assertTrue(stop_key in get_kvs)
        self.assertEqual(get_kvs[stop_key], stop_value)

        # True False
        opt.start_inclusive = True
        opt.stop_inclusive = False
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   start_key,
                                                   stop_key,
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), 1)
        self.assertTrue(start_key in get_kvs)
        self.assertEqual(get_kvs[start_key], start_value)

        # True True
        opt.start_inclusive = True
        opt.stop_inclusive = True
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   start_key,
                                                   stop_key,
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), 2)
        self.assertTrue(start_key in get_kvs)
        self.assertEqual(get_kvs[start_key], start_value)
        self.assertTrue(stop_key in get_kvs)
        self.assertEqual(get_kvs[stop_key], stop_value)

    @inlineCallbacks
    def test_multi_get_opt_no_value_ok(self):
        count = 50
        rand_key = uuid.uuid1().hex
        kvs = {self.TEST_SKEY + rand_key + "_" + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        opt = MultiGetOptions()
        opt.no_value = True
        (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                   self.TEST_SKEY + rand_key + '_0',
                                                   self.TEST_SKEY + rand_key + '_' + '9'*len(str(count)),
                                                   opt,
                                                   500)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(kvs))
        for k in kvs.keys():
            self.assertTrue(k in get_kvs)

    @inlineCallbacks
    def test_multi_get_opt_reverse_ok(self):
        count = 50
        steps = 5
        self.assertTrue(count % steps == 0)
        split_count = count / 5
        rand_key = uuid.uuid1().hex
        ks = sorted({self.TEST_SKEY + rand_key + "_" + str(x) for x in range(count)})
        kvs = {self.TEST_SKEY + rand_key + "_" + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        stop_key = self.TEST_SKEY + rand_key + '_' + '9'*len(str(count))
        opt = MultiGetOptions()
        opt.reverse = True
        for step in range(steps):
            (rc, get_kvs) = yield self.c.multi_get_opt(self.TEST_HKEY,
                                                       self.TEST_SKEY + rand_key + '_0',
                                                       stop_key,
                                                       opt,
                                                       split_count)
            if step == steps - 1:
                self.assertEqual(rc, error_types.ERR_OK.value)
            else:
                self.assertEqual(rc, error_types.ERR_INCOMPLETE_DATA.value)
            self.assertEqual(len(get_kvs), split_count)
            for _ in range(split_count):
                k = ks.pop()
                self.assertTrue(k in get_kvs)
                self.assertEqual(get_kvs[k], kvs[k])
                stop_key = k

    @inlineCallbacks
    def test_multi_get_1by1_ok(self):
        count = 5
        ks = {self.TEST_SKEY + str(x) for x in range(count)}
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        get_count = 0
        while ks:
            (rc, get_kvs) = yield self.c.multi_get(self.TEST_HKEY, ks, 1)
            if rc == error_types.ERR_INCOMPLETE_DATA.value\
               or rc == error_types.ERR_OK.value:
                for (k, v) in get_kvs.items():
                    get_count += 1
                    self.assertIn(k, ks)
                    ks.remove(k)
                    self.assertIn(k, kvs)
                    self.assertEqual(v, kvs[k])

        self.assertEqual(get_count, len(kvs))

    @inlineCallbacks
    def test_multi_del_ok(self):
        count = 50
        ks = {self.TEST_SKEY + str(x) for x in range(count)}

        (rc, del_count) = yield self.c.multi_del(self.TEST_HKEY, ks)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(del_count, len(ks))

    @inlineCallbacks
    def test_multi_get_part_ok(self):
        count = 50
        ks = {self.TEST_SKEY + str(x) for x in range(count/2)}
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, get_kvs) = yield self.c.multi_get(self.TEST_HKEY, ks)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(ks))
        for (k, v) in get_kvs.items():
            self.assertIn(k, ks)
            self.assertIn(k, kvs)
            self.assertEqual(v, kvs[k])

    @inlineCallbacks
    def test_multi_get_more_ok(self):
        count = 50
        ks = {self.TEST_SKEY + str(x) for x in range(count)}
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count/2)}

        (rc, del_count) = yield self.c.multi_del(self.TEST_HKEY, ks)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(del_count, len(ks))

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, get_kvs) = yield self.c.multi_get(self.TEST_HKEY, ks)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_kvs), len(kvs))
        for (k, v) in get_kvs.items():
            self.assertIn(k, ks)
            self.assertIn(k, kvs)
            self.assertEqual(v, kvs[k])

    @inlineCallbacks
    def test_sort_key_count_ok(self):
        rand_key = uuid.uuid1().hex
        (rc, count) = yield self.c.sort_key_count(self.TEST_HKEY + rand_key)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(count, 0)

    @inlineCallbacks
    def test_get_sort_keys_none(self):
        rand_key = uuid.uuid1().hex
        (rc, get_ks) = yield self.c.get_sort_keys(self.TEST_HKEY + rand_key)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_ks), 0)

    @inlineCallbacks
    def test_get_sort_keys_ok(self):
        count = 50
        rand_key = uuid.uuid1().hex
        ks = {self.TEST_SKEY + str(x) for x in range(count)}
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        (rc, del_count) = yield self.c.multi_del(self.TEST_HKEY + rand_key, ks)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(del_count, len(ks))

        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY + rand_key, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        (rc, get_ks) = yield self.c.get_sort_keys(self.TEST_HKEY + rand_key)
        self.assertEqual(rc, error_types.ERR_OK.value)
        self.assertEqual(len(get_ks), count)

    @inlineCallbacks
    def test_scan_none(self):
        rand_key = uuid.uuid1().hex
        o = ScanOptions()
        s = self.c.get_scanner(self.TEST_HKEY + rand_key, b'\x00\x00', b'\xFF\xFF', o)
        ret = yield s.get_next()
        self.assertEqual(ret, None)
        s.close()

    @inlineCallbacks
    def test_scan_all(self):
        count = 50
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        rand_key = uuid.uuid1().hex
        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY + rand_key, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        o = ScanOptions()
        s = self.c.get_scanner(self.TEST_HKEY + rand_key, b'\x00\x00', b'\xFF\xFF', o)
        get_count = 0
        last_sk = None
        while True:
            hk_sk_v = yield s.get_next()
            if hk_sk_v is None:
                break
            else:
                get_count += 1
                hk = hk_sk_v[0][0]
                sk = hk_sk_v[0][1]
                v = hk_sk_v[1]

                self.assertNotEqual(sk, None)
                if last_sk is not None:
                    self.assertLess(last_sk, sk)
                last_sk = sk

                self.assertEqual(hk, self.TEST_HKEY + rand_key)
                self.assertIn(sk, kvs)
                self.assertEqual(v, kvs[sk])
        s.close()
        self.assertEqual(count, get_count)

    @inlineCallbacks
    def test_scan_part(self):
        count = 50
        self.assertLess(2, count)
        count_len = len(str(count))
        kvs = {self.TEST_SKEY + str(x).zfill(count_len): self.TEST_VALUE + str(x) for x in range(count)}
        sub_kvs = {self.TEST_SKEY + str(x).zfill(count_len): self.TEST_VALUE + str(x) for x in range(1, count-1)}

        rand_key = uuid.uuid1().hex
        (ret, ign) = yield self.c.multi_set(self.TEST_HKEY + rand_key, kvs)
        self.assertEqual(ret, error_types.ERR_OK.value)

        o = ScanOptions()
        s = self.c.get_scanner(self.TEST_HKEY + rand_key,
                               self.TEST_SKEY + str(1).zfill(count_len),
                               self.TEST_SKEY + str(count-1).zfill(count_len),
                               o)
        get_count = 0
        last_sk = None
        while True:
            hk_sk_v = yield s.get_next()
            if hk_sk_v is None:
                break
            else:
                get_count += 1
                hk = hk_sk_v[0][0]
                sk = hk_sk_v[0][1]
                v = hk_sk_v[1]

                self.assertNotEqual(sk, None)
                if last_sk is not None:
                    self.assertLess(last_sk, sk)
                last_sk = sk

                self.assertEqual(hk, self.TEST_HKEY + rand_key)
                self.assertIn(sk, sub_kvs)
                self.assertEqual(v, sub_kvs[sk])
        s.close()
        self.assertEqual(len(sub_kvs), get_count)

    @inlineCallbacks
    def test_unordered_scan_none(self):
        split_count = 5
        o = ScanOptions()
        scanners = self.c.get_unordered_scanners(split_count, o)
        self.assertEqual(len(scanners), split_count)
        for scanner in scanners:
            while True:
                hk_sk_v = yield scanner.get_next()
                # self.assertEqual(hk_sk_v, None)                   # TODO there maybe some remain data
                if hk_sk_v is None:
                    break
            scanner.close()

    @inlineCallbacks
    def test_unordered_scan_all(self):
        hkey_count = 20
        count = 50
        rand_key = uuid.uuid1().hex
        hks = {self.TEST_HKEY + rand_key + '_' + str(i) for i in range(hkey_count)}
        kvs = {self.TEST_SKEY + str(x): self.TEST_VALUE + str(x) for x in range(count)}

        for hk in hks:
            (ret, ign) = yield self.c.multi_set(hk, kvs)
            self.assertEqual(ret, error_types.ERR_OK.value)

        split_count = 5
        o = ScanOptions()
        scanners = self.c.get_unordered_scanners(split_count, o)
        self.assertEqual(len(scanners), split_count)

        get_hk_count = 0
        for scanner in scanners:
            last_hk = None
            last_sk = None
            hk_found = False
            get_count = 0
            while True:
                hk_sk_v = yield scanner.get_next()
                if hk_sk_v is None:
                    break

                hk = hk_sk_v[0][0]
                sk = hk_sk_v[0][1]
                v = hk_sk_v[1]

                self.assertNotEqual(hk, None)
                if last_hk is None:
                    last_hk = hk
                    get_hk_count += 1

                self.assertNotEqual(sk, None)
                if hk == last_hk:
                    get_count += 1
                    if last_sk is not None:
                        self.assertLess(last_sk, sk)
                else:
                    get_hk_count += 1
                    if hk_found:
                        self.assertEqual(count, get_count)
                    get_count = 1
                    last_hk = hk
                last_sk = sk

                if hk in hks:
                    hk_found = True
                    self.assertIn(sk, kvs)
                    self.assertEqual(v, kvs[sk])
                else:
                    hk_found = False

            scanner.close()

        self.assertLessEqual(hkey_count, get_hk_count)



#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
import pypegasus.base.ttypes


from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None



class partition_configuration:
  """
  Attributes:
   - pid
   - ballot
   - max_replica_count
   - primary
   - secondaries
   - last_drops
   - last_committed_decree
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'pid', (pypegasus.base.ttypes.gpid, pypegasus.base.ttypes.gpid.thrift_spec), None, ), # 1
    (2, TType.I64, 'ballot', None, None, ), # 2
    (3, TType.I32, 'max_replica_count', None, None, ), # 3
    (4, TType.STRUCT, 'primary', (pypegasus.base.ttypes.rpc_address, pypegasus.base.ttypes.rpc_address.thrift_spec), None, ), # 4
    (5, TType.LIST, 'secondaries', (TType.STRUCT,(pypegasus.base.ttypes.rpc_address, pypegasus.base.ttypes.rpc_address.thrift_spec)), None, ), # 5
    (6, TType.LIST, 'last_drops', (TType.STRUCT,(pypegasus.base.ttypes.rpc_address, pypegasus.base.ttypes.rpc_address.thrift_spec)), None, ), # 6
    (7, TType.I64, 'last_committed_decree', None, None, ), # 7
  )

  def __init__(self, pid=None, ballot=None, max_replica_count=None, primary=None, secondaries=None, last_drops=None, last_committed_decree=None,):
    self.pid = pid
    self.ballot = ballot
    self.max_replica_count = max_replica_count
    self.primary = primary
    self.secondaries = secondaries
    self.last_drops = last_drops
    self.last_committed_decree = last_committed_decree

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.pid = pypegasus.base.ttypes.gpid()
          self.pid.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.ballot = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I32:
          self.max_replica_count = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRUCT:
          self.primary = pypegasus.base.ttypes.rpc_address()
          self.primary.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.LIST:
          self.secondaries = []
          (_etype3, _size0) = iprot.readListBegin()
          for _i4 in xrange(_size0):
            _elem5 = pypegasus.base.ttypes.rpc_address()
            _elem5.read(iprot)
            self.secondaries.append(_elem5)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 6:
        if ftype == TType.LIST:
          self.last_drops = []
          (_etype9, _size6) = iprot.readListBegin()
          for _i10 in xrange(_size6):
            _elem11 = pypegasus.base.ttypes.rpc_address()
            _elem11.read(iprot)
            self.last_drops.append(_elem11)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 7:
        if ftype == TType.I64:
          self.last_committed_decree = iprot.readI64()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('partition_configuration')
    if self.pid is not None:
      oprot.writeFieldBegin('pid', TType.STRUCT, 1)
      self.pid.write(oprot)
      oprot.writeFieldEnd()
    if self.ballot is not None:
      oprot.writeFieldBegin('ballot', TType.I64, 2)
      oprot.writeI64(self.ballot)
      oprot.writeFieldEnd()
    if self.max_replica_count is not None:
      oprot.writeFieldBegin('max_replica_count', TType.I32, 3)
      oprot.writeI32(self.max_replica_count)
      oprot.writeFieldEnd()
    if self.primary is not None:
      oprot.writeFieldBegin('primary', TType.STRUCT, 4)
      self.primary.write(oprot)
      oprot.writeFieldEnd()
    if self.secondaries is not None:
      oprot.writeFieldBegin('secondaries', TType.LIST, 5)
      oprot.writeListBegin(TType.STRUCT, len(self.secondaries))
      for iter12 in self.secondaries:
        iter12.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.last_drops is not None:
      oprot.writeFieldBegin('last_drops', TType.LIST, 6)
      oprot.writeListBegin(TType.STRUCT, len(self.last_drops))
      for iter13 in self.last_drops:
        iter13.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.last_committed_decree is not None:
      oprot.writeFieldBegin('last_committed_decree', TType.I64, 7)
      oprot.writeI64(self.last_committed_decree)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.pid)
    value = (value * 31) ^ hash(self.ballot)
    value = (value * 31) ^ hash(self.max_replica_count)
    value = (value * 31) ^ hash(self.primary)
    value = (value * 31) ^ hash(self.secondaries)
    value = (value * 31) ^ hash(self.last_drops)
    value = (value * 31) ^ hash(self.last_committed_decree)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class query_cfg_request:
  """
  Attributes:
   - app_name
   - partition_indices
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'app_name', None, None, ), # 1
    (2, TType.LIST, 'partition_indices', (TType.I32,None), None, ), # 2
  )

  def __init__(self, app_name=None, partition_indices=None,):
    self.app_name = app_name
    self.partition_indices = partition_indices

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.app_name = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.LIST:
          self.partition_indices = []
          (_etype17, _size14) = iprot.readListBegin()
          for _i18 in xrange(_size14):
            _elem19 = iprot.readI32()
            self.partition_indices.append(_elem19)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('query_cfg_request')
    if self.app_name is not None:
      oprot.writeFieldBegin('app_name', TType.STRING, 1)
      oprot.writeString(self.app_name)
      oprot.writeFieldEnd()
    if self.partition_indices is not None:
      oprot.writeFieldBegin('partition_indices', TType.LIST, 2)
      oprot.writeListBegin(TType.I32, len(self.partition_indices))
      for iter20 in self.partition_indices:
        oprot.writeI32(iter20)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.app_name)
    value = (value * 31) ^ hash(self.partition_indices)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class query_cfg_response:
  """
  Attributes:
   - err
   - app_id
   - partition_count
   - is_stateful
   - partitions
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'err', (pypegasus.base.ttypes.error_code, pypegasus.base.ttypes.error_code.thrift_spec), None, ), # 1
    (2, TType.I32, 'app_id', None, None, ), # 2
    (3, TType.I32, 'partition_count', None, None, ), # 3
    (4, TType.BOOL, 'is_stateful', None, None, ), # 4
    (5, TType.LIST, 'partitions', (TType.STRUCT,(partition_configuration, partition_configuration.thrift_spec)), None, ), # 5
  )

  def __init__(self, err=None, app_id=None, partition_count=None, is_stateful=None, partitions=None,):
    self.err = err
    self.app_id = app_id
    self.partition_count = partition_count
    self.is_stateful = is_stateful
    self.partitions = partitions

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.err = pypegasus.base.ttypes.error_code()
          self.err.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I32:
          self.app_id = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I32:
          self.partition_count = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.BOOL:
          self.is_stateful = iprot.readBool()
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.LIST:
          self.partitions = []
          (_etype24, _size21) = iprot.readListBegin()
          for _i25 in xrange(_size21):
            _elem26 = partition_configuration()
            _elem26.read(iprot)
            self.partitions.append(_elem26)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('query_cfg_response')
    if self.err is not None:
      oprot.writeFieldBegin('err', TType.STRUCT, 1)
      self.err.write(oprot)
      oprot.writeFieldEnd()
    if self.app_id is not None:
      oprot.writeFieldBegin('app_id', TType.I32, 2)
      oprot.writeI32(self.app_id)
      oprot.writeFieldEnd()
    if self.partition_count is not None:
      oprot.writeFieldBegin('partition_count', TType.I32, 3)
      oprot.writeI32(self.partition_count)
      oprot.writeFieldEnd()
    if self.is_stateful is not None:
      oprot.writeFieldBegin('is_stateful', TType.BOOL, 4)
      oprot.writeBool(self.is_stateful)
      oprot.writeFieldEnd()
    if self.partitions is not None:
      oprot.writeFieldBegin('partitions', TType.LIST, 5)
      oprot.writeListBegin(TType.STRUCT, len(self.partitions))
      for iter27 in self.partitions:
        iter27.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.err)
    value = (value * 31) ^ hash(self.app_id)
    value = (value * 31) ^ hash(self.partition_count)
    value = (value * 31) ^ hash(self.is_stateful)
    value = (value * 31) ^ hash(self.partitions)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class request_meta:
  """
  Attributes:
   - app_id
   - partition_index
   - client_timeout
   - partition_hash
   - is_backup_request
  """

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'app_id', None, None, ), # 1
    (2, TType.I32, 'partition_index', None, None, ), # 2
    (3, TType.I32, 'client_timeout', None, None, ), # 3
    (4, TType.I64, 'partition_hash', None, None, ), # 4
    (5, TType.BOOL, 'is_backup_request', None, None, ), # 5
  )

  def __init__(self, app_id=None, partition_index=None, client_timeout=None, partition_hash=None, is_backup_request=None,):
    self.app_id = app_id
    self.partition_index = partition_index
    self.client_timeout = client_timeout
    self.partition_hash = partition_hash
    self.is_backup_request = is_backup_request

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I32:
          self.app_id = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I32:
          self.partition_index = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I32:
          self.client_timeout = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.I64:
          self.partition_hash = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.BOOL:
          self.is_backup_request = iprot.readBool()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('request_meta')
    if self.app_id is not None:
      oprot.writeFieldBegin('app_id', TType.I32, 1)
      oprot.writeI32(self.app_id)
      oprot.writeFieldEnd()
    if self.partition_index is not None:
      oprot.writeFieldBegin('partition_index', TType.I32, 2)
      oprot.writeI32(self.partition_index)
      oprot.writeFieldEnd()
    if self.client_timeout is not None:
      oprot.writeFieldBegin('client_timeout', TType.I32, 3)
      oprot.writeI32(self.client_timeout)
      oprot.writeFieldEnd()
    if self.partition_hash is not None:
      oprot.writeFieldBegin('partition_hash', TType.I64, 4)
      oprot.writeI64(self.partition_hash)
      oprot.writeFieldEnd()
    if self.is_backup_request is not None:
      oprot.writeFieldBegin('is_backup_request', TType.BOOL, 5)
      oprot.writeBool(self.is_backup_request)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.app_id)
    value = (value * 31) ^ hash(self.partition_index)
    value = (value * 31) ^ hash(self.client_timeout)
    value = (value * 31) ^ hash(self.partition_hash)
    value = (value * 31) ^ hash(self.is_backup_request)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)
